package org.dimx;

import cleoni.adv.*;
import java.util.Vector;
/**
 *
 * @author Carl Nagle
 */
public class AdvancedLanguageParser extends LanguageParser 
                                    implements IWorldInitProcessor{

    public static final String NOUN_LC = "noun";
    public static final String PRONOUN_LC = "pronoun";
    public static final String PLURAL_LC = "plural";
    public static final String ADJS_LC = "adjs";
    public static final String PREADJ_LC = "preadj";
    public static final String TYPE_LC = "type";
    public static final String HS_MARK = "$HS";
    public static final String HE_MARK = "$HE";
    public static final String HS_KEY = "hilite_start";
    public static final String HE_KEY = "hilite_end";
    
    public String HS_STRING = "<i CLASS='iconlabel'>";
    public String HE_STRING = "</i>";
    
    /** "alpmsgsFile" 
     * value in WORLD ATTRLIST identifying the file to use for custom messages:
     * Ex: ATTRLIST alpmsgsFile=CSI.msgs,alpmsgs=csimsg
     **/
    public static final String ALPMSGSFILEID = "alpmsgsFile";
    /** "alpmsgs" 
     * value in WORLD ATTRLIST identifying the World varname to use in Dimx Scripts
     * Ex: ATTRLIST alpmsgsFile=CSI.msgs,alpmsgs=csimsg
     * Scripts reference the custom messages as a SET with this name.
     * 
     **/
    public static final String ALPMSGSVARID = "alpmsgs";
    Dict cmsgs = new Dict();
    
    String alpmsgsvar = ALPMSGSVARID;
    
    public Dict verbs = new Dict();
    //alt say syntax: tell someone something, instead of say something to someone
    public Dict say2 = new Dict(); //tell someone something
    //alt use syntax: unlock door with key, instead of use key on door
    public Dict use2 = new Dict(); 
    public Dict cmdAliases = new Dict();    //key=alias;value=cmd
    public Dict preAdjectives = new Dict(); //pa=pa  (the, a, an, etc)
    public Dict preAdjNouns = new Dict();   //key=noun;value=preadj
    public Dict adjectives = new Dict();    //aj=aj
    public Dict nouns = new Dict();         //nn=nn
    public Dict plurals = new Dict();       //pl=pl
    public Dict pluralNouns = new Dict();   //key=plural,value=noun
    public Dict nounPlurals = new Dict();   //key=noun,value=plural
    public Dict pronouns = new Dict();      //pn=pn
    public Dict pronounNouns = new Dict();  //key=noun,value=pronoun
    public Dict withs = new Dict();         //w=w
    public Dict ands = new Dict();          //a=a

    protected String getTxt(String key){
        return ((Token)cmsgs.get(key)).strVal();
    }
    
    /**
     * The workhorse routine to parse user input to execute game commands
     * @param server
     * @param world
     * @param cmd
     * @param arg1
     * @param arg2
     * @param player
     * @return NULL to allow normal SAY processing to continue; 
     *         Vector{new cmd,arg1,arg2} to bypass normal SAY processing.
     * @override org.dimx.LanguageParser#parseCommands(multiplayer,World,String,String,String,Player)
     */
    Vector parseCommands (multiplayer server, World world, String cmd, String arg1, String arg2, Player player) throws DimxException{
        
        logmsg(world,"ALP: processing input: "+ arg1);
        initLocalizedData(world);//will only do so once per world
        String arg1_lc = arg1.toLowerCase();
        player.properties.put("alpinput", new Token(arg1_lc));
        Vector utils = clearCmdVector();
        utils.set(1, arg1); // the say input
        String [] words = arg1.split(space); //TODO: watch for punctuation (,.!?, etc.)
        int i = 0;
        String verb = null;
        String alias = null;
        boolean aliasValid = false;
        int predStart = -1;
        String objtext = null;
        String predtext = null;
        player.getClient().cmdNewmsg(); // clear the txtBox on next refresh
        while(i < words.length){            
            verb = words[i++].toLowerCase();
            alias = cmdAliases.getS(verb);
            aliasValid = !(alias == null);
            world.debug("ALP: word="+ verb);
            if(verb.equalsIgnoreCase(getTxt("command_go"))) continue;//go
            if(aliasValid && alias.equalsIgnoreCase(getTxt("command_go"))) continue;
            if(!aliasValid){
                if(verb.equals(N_LC)||verb.equalsIgnoreCase(world.msgs.msg[167])){
                    utils.set(1, N_UC);
                    return cmdGO(server, world, utils, player);
                }
                if(verb.equals(S_LC)||verb.equalsIgnoreCase(world.msgs.msg[168])) {
                    utils.set(1, S_UC);
                    return cmdGO(server, world, utils, player);
                }
                if(verb.equals(E_LC)||verb.equalsIgnoreCase(world.msgs.msg[169])){
                    utils.set(1, E_UC);
                    return cmdGO(server, world, utils, player);
                }
                if(verb.equals(W_LC)||verb.equalsIgnoreCase(world.msgs.msg[170])){
                    utils.set(1, W_UC);
                    return cmdGO(server, world, utils, player);
                }
                if(verb.equals(U_LC)||verb.equalsIgnoreCase(world.msgs.msg[171])){
                    utils.set(1, U_UC);
                    return cmdGO(server, world, utils, player);
                }
                if(verb.equals(D_LC)||verb.equalsIgnoreCase(world.msgs.msg[172])){
                    utils.set(1, D_UC);
                    return cmdGO(server, world, utils, player);
                }
                if(verb.equals(TL_LC)||verb.equalsIgnoreCase(getTxt("command_left"))){
                    utils.set(1, TL_LC);
                    return cmdTURN(server, world, utils, player);
                }
                if(verb.equals(TR_LC)||verb.equalsIgnoreCase(getTxt("command_right"))){
                    utils.set(1, TR_LC);
                    return cmdTURN(server, world, utils, player);
                }
                if(verb.equals(TURN_LC)||world.msgs.msg[173].toLowerCase().startsWith(verb)) /* turn */
                    return cmdTURN(server, world, utils, player);            
                if(verb.equals(L_LC)) /* l || Look */
                    return cmdLOOK(server, world, utils, player);
                
                //engine commands /HELP, /SAVE, /SAVE EXIT, /DISCONNECT
                if(verb.equalsIgnoreCase(getTxt("command_help"))||verb.equalsIgnoreCase(SLASH_HELP))
                    return cmdHELP(server, world, utils, player);
                if(verb.equalsIgnoreCase(getTxt("command_save"))||verb.equalsIgnoreCase(SLASH_SAVE))
                    return cmdSAVE(server, world, utils, player);
                if(verb.equalsIgnoreCase(getTxt("command_logout"))||verb.equalsIgnoreCase(SLASH_DISCONNECT))
                    return cmdDISCONNECT(server, world, utils, player);
                //if no alias then this particular word is not a supported verb or cmd
                continue;
            }
            // commands likely to have aliases
            if(alias.equalsIgnoreCase(getTxt("command_look"))||alias.equalsIgnoreCase(world.msgs.cmd[5]))
                return cmdLOOK(server, world, utils, player);
            if(alias.equalsIgnoreCase(getTxt("command_pick"))||alias.equalsIgnoreCase(world.msgs.cmd[8]))
                return cmdPICK(server, world, utils, player);
            if(alias.equalsIgnoreCase(getTxt("command_drop"))||alias.equalsIgnoreCase(world.msgs.cmd[9]))
                return cmdDROP(server, world, utils, player);
            if(alias.equalsIgnoreCase(getTxt("command_search"))||alias.equalsIgnoreCase(world.msgs.cmd[10]))
                return cmdSEARCH(server, world, utils, player);
            if(alias.equalsIgnoreCase(getTxt("command_open"))||alias.equalsIgnoreCase(world.msgs.cmd[6]))
                return cmdOPEN(server, world, utils, player);
            if(alias.equalsIgnoreCase(getTxt("command_close"))||alias.equalsIgnoreCase(world.msgs.cmd[7]))
                return cmdCLOSE(server, world, utils, player);
            if(alias.equalsIgnoreCase(getTxt("command_hide"))||alias.equalsIgnoreCase(world.msgs.cmd[11]))
                return cmdHIDE(server, world, utils, player);
            if(alias.equalsIgnoreCase(getTxt("command_enter"))||alias.equalsIgnoreCase(world.msgs.cmd[22]))
                return cmdENTER(server, world, utils, player);
            if(alias.equalsIgnoreCase(getTxt("command_exit"))||alias.equalsIgnoreCase(world.msgs.cmd[23]))
                return cmdEXIT(server, world, utils, player);
            if(alias.equalsIgnoreCase(getTxt("command_reverse"))||alias.equalsIgnoreCase(world.msgs.cmd[25]))
                return cmdREVERSE(server, world, utils, player);

            //tell minion something
            if(say2.containsKey(verb)){
                predStart = arg1_lc.indexOf(verb + space);                
                if (predStart == -1) continue; //end of sentence
                try { return cmdSAY2(server, world, utils, player, arg1.substring(predStart + verb.length()+1));}
                catch(IndexOutOfBoundsException iob){ continue; }
            }
            
            // handle 2-object commands
            predStart = indexAfterPredicate(world, arg1_lc);
            objtext = (predStart == -1) ? arg1 : arg1.substring(0,predStart);
            predtext = (predStart == -1) ? "" : arg1.substring(predStart);

            //say something to minion
            if(alias.equalsIgnoreCase(getTxt("command_say"))||alias.equalsIgnoreCase(world.msgs.cmd[14]))
                return cmdSAYTO(server, world, utils, player);
            if(alias.equalsIgnoreCase(getTxt("command_put"))||alias.equalsIgnoreCase(world.msgs.cmd[12])) {
                if(predStart==-1) return badPUT(world, player, arg1);
                return cmdPUT(server, world, utils, player, objtext, predtext);
            }
            if(alias.equalsIgnoreCase(getTxt("command_give"))||alias.equalsIgnoreCase(world.msgs.cmd[13])) {
                if(predStart==-1) return badGIVE(world, player, arg1);
                return cmdGIVE(server, world, utils, player, objtext, predtext);
            }
            //unlock door with key
            if (use2.containsKey(verb)){
                return cmdUSE(server, world, utils, player, predtext, objtext);
            }
            //use key on door
            if(alias.equalsIgnoreCase(getTxt("command_use"))||alias.equalsIgnoreCase(world.msgs.cmd[3])) { 
                return cmdUSE(server, world, utils, player, objtext, predtext);
            }
        }
        return null;    
    }

    protected Vector cmdSAY2(multiplayer server, World world, Vector utils, Player player, String objtext){

        logmsg(world,"ALP SAY2 '"+ objtext +"'");
        
        Vector info = getPlayerContainerCharacter(world, player, objtext);
        Vector ids   = (Vector)info.elementAt(0);
        Vector texts = (Vector)info.elementAt(1);
        if((ids.size() == 0) || (texts.size() == 0)) {// not found
            logmsg(world,"ALP SAY2 target does not exist. Reverting to normal say output.");
            utils.set(0, SAY_LC);
            utils.set(1, objtext);
            utils.set(2, "");
            return utils;
        }        
        utils.set(0, SAY_LC);
        utils.set(1, (String)texts.elementAt(0));
        utils.set(2, (String)ids.elementAt(0));
        // new CHAT version of Say
        return cmdSAY(server, world, utils, player);
    }

    /**
     * Attempt to match the input_lc with the Name of an AdvObject in the items Dict.
     * Will attempt to id items in open container items, too.
     * if we match an adj then ALL matches must match on adj, too
     * if we match name then ALL matches must match on name, too
     * @param world
     * @param input_lc
     * @param items
     * @param namematched if we match name then ALL matches must match on name, too
     * @param adjmatched if we match an adj then ALL matches must match on adj, too
     * @return 0 or more item Ids who's full NAME can be found in text_lc
     */
    protected String[] getMatchingIds(World world, String input_lc, Dict items, boolean namematched, boolean adjmatched){
        Vector ids = new Vector();
        if(items.size()==0) return new String[0];        
        String id = null;
        AdvObject obj = null;
        String oname = null;
        Dict contents = null;
        Token token = null;
        String rawwords = null;
        String[] words = null;
        String word = null;
        String input_adjs = null; //substring BEFORE noun
        int tindex = -1;
        logmsg(world, "ALP processing for ids: "+ input_lc);
        for(int i = 0;i < items.size(); i++){
            obj = (AdvObject)items.elementAt(i);
            id = obj.getId();
            oname = obj.name.strVal().toLowerCase();
            //try match on the item's complete name
            if(input_lc.indexOf(oname)>= 0){
                logmsg(world, "ALP matched on name: "+ oname);
                if(!namematched) ids.clear(); //delete matches that aren't as strict
                namematched = true;
                ids.add(id);
            }
            // try to match on plural, noun, pronoun, adjs, etc...
            else if(! namematched){
                logmsg(world, "ALP seeking plurals, nouns, and pronouns for "+ oname);
                tindex = -1;
                //this order is important: plural, noun, pronoun
                // plural?
                token = obj.varGet(PLURAL_LC);
                if(!(token == null)){
                    word = token.strVal().toLowerCase();
                    if (word.length()> 0)
                        tindex = input_lc.indexOf(word);
                }
                if (tindex == -1){
                    token = obj.varGet(NOUN_LC);
                    if(!(token==null)){
                        word = token.strVal().toLowerCase();
                        if (word.length()> 0)
                            tindex = input_lc.indexOf(word);
                        }
                }
                // pronouns? possibly multiple pronouns (item thing, etc)
                if(tindex == -1){
                    token = obj.varGet(PRONOUN_LC);
                    if(!(token == null)){
                        rawwords = token.strVal().toLowerCase();
                        words = rawwords.split(space);
                        for(int w =0;(w < words.length)&&(tindex < 0);w++){
                            word = words[w];
                            if (word.length()> 0)
                                tindex = input_lc.indexOf(words[w].toLowerCase());
                        }
                    }
                }

                if (tindex == -1){
                    logmsg(world, "ALP did NOT match on plurals, nouns, and pronouns for "+ oname);
                    //not match, but check its contents at bottom
                }else if (tindex < 2 ){ // no room for adjs?
                    if(!(adjmatched)){
                        logmsg(world, "ALP matched plural, noun, or pronoun: "+ oname +" with "+ word);
                        ids.add(id);                        
                    }
                }else{
                    // if matched, make adjs substring to search--everything before the noun
                    input_adjs = input_lc.substring(0, tindex);
                    tindex = -1;
                    logmsg(world, "ALP try matching on adjs: "+ input_adjs);
                    // search for adjs
                    token = obj.varGet(ADJS_LC);
                    if(!(token==null)){
                        rawwords = token.strVal().toLowerCase();
                        words = rawwords.split(space);
                        for(int w =0;(w < words.length)&&(tindex < 0);w++){
                            word = words[w];
                            if (word.length()> 0)
                                tindex = input_lc.indexOf(words[w].toLowerCase());
                        }
                        if(tindex > -1){
                            if(!adjmatched) ids.clear();
                            adjmatched = true;
                            ids.add(id);
                            logmsg(world, "ALP matched on one or more adjs: "+ token.strVal());
                        }else if(!adjmatched) {
                            ids.add(id);
                        }
                    }else if(!adjmatched){
                        ids.add(id);
                    }
                }
            }
            // check for contents of open containers
            contents = obj.getContents(true);
            if(contents != null){
                String[] temp = getMatchingIds(world, input_lc, contents, namematched, adjmatched);
                for(int t = 0; t < temp.length;t++) ids.add(temp[t]);
            }
        }// for loop iteration
        return (String[])ids.toArray(new String[ids.size()]);
    }
    
    /**
     * Attempt to match the input_lc with the Name of an AdvObject in the items Dict.
     * Will attempt to id items in open container items, too.
     * @param world
     * @param input_lc
     * @param items
     * @return 0 or more item Ids who's full NAME can be found in text_lc
     * @override org.dimx.LanguageParser#getMatchingIds(World,String,Dict)
     */
    protected String[] getMatchingIds(World world, String input_lc, Dict items){
        return getMatchingIds(world, input_lc, items, false, false);
    }
    
    /**
     * 
     * @param world
     * @param player
     * @param objtext
     * @return Vector(0:Vector(ids),1:Vector(texts))
     */
    Vector getPlayerContainerCharacter(World world, Player player, String objtext){
        Vector result = new Vector(2);
        Vector ids = new Vector();
        Vector texts = new Vector();
        result.add(ids);
        result.add(texts);
        
        String objtext_lc = objtext.toLowerCase();
        Dict items = player.container.getContents(false);
        if(items.size()==0) return result;        

        String id = null;
        AdvObject obj = null;
        String oname = null;
        String[]names = null;
        Dict contents = null;
        int index = -1;
        for(int i = 0;i < items.size(); i++){
            obj = (AdvObject)items.elementAt(i);
            id = obj.getId();
            oname = obj.name.strVal().toLowerCase();
            names = oname.split(space);
            //currently we only match on the item's name
            index = objtext_lc.indexOf(oname+space);
            if(index >= 0){
                try{
                    texts.add(objtext.substring(index + oname.length()+1).trim());
                    ids.add(id);
                }catch(IndexOutOfBoundsException iob){}
                continue;
            }
            if(names.length==1) continue;
            //try partial match or out of order
            for(int n = 0;n<names.length;n++){
                index = objtext_lc.indexOf(names[n]+space);
                if(index >= 0){
                    try{
                        texts.add(objtext.substring(index + names[n].length()+1).trim());
                        ids.add(id);
                    }catch(IndexOutOfBoundsException iob){}
                    break;
                }
            }
        }
        return result;
    }
    
    /**
     * This API is subject to change until it is fully developed.
     * @param server
     * @param loader
     * @param world
     * @param systemDir
     * @return boolean true to allow normal world.init processing to continue;
     *         false to bypass normal world.init processing.
     */
    public boolean processPreWorldInit(multiplayer server, WorldLoader loader, World world, String systemDir){
        return true; //do nothing
    }
    
    /**
     * Perform custom initialization and allow normal world.init to finalize.
     * This API is subject to change until it is fully developed.
     * @param server
     * @param loader
     * @param world
     * @param systemDir
     * @return boolean true to allow remaining init processing to continue;
     *         false to bypass remaining init processing.
     */
    public boolean processPostWorldInit(multiplayer server, WorldLoader loader, World world, String systemDir) {
        logmsg(world, "ALP Custom post-init processing complete.");
        //if(true) return true;
        AdvObject obj = null;
        Token adjs = null;
        Token preadj = null;
        Token noun = null;
        Token pronoun = null;
        Token plural = null;
        
        Dict _objs = world.getObjects();        
        String customsgsFile = world.varGet(ALPMSGSFILEID).strVal();
        String alpmsgsvarname = ALPMSGSVARID;
        if(world.varExists(ALPMSGSVARID))
            alpmsgsvarname = world.varGet(ALPMSGSVARID).strVal();        
        if(customsgsFile.length()>0) {
            customsgsFile = systemDir + customsgsFile;
            logmsg(world, "ALP attempting load of custom messages from: "+ customsgsFile);
            try {
                java.util.Properties msgs = new java.util.Properties();
                msgs.load(new java.io.FileInputStream(customsgsFile));
                java.util.Enumeration keys = msgs.keys();
                if(keys.hasMoreElements()){
                    Dict _msgs = new Dict();
                    String key = null;
                    String val = null;
                    while(keys.hasMoreElements()){
                        key = (String)keys.nextElement();
                        val = msgs.getProperty(key);
                        _msgs.put(key, new Token(val));
                    }
                    Token t = new Token(_msgs);
                    world.varGet(alpmsgsvarname, Const.GETREF).assign(t, world);
                    cmsgs = _msgs;
                    HS_STRING = msgs.getProperty(HS_KEY, HS_STRING);
                    HE_STRING = msgs.getProperty(HE_KEY, HE_STRING);
                }
            } catch (java.io.FileNotFoundException e) {
                logmsg(world, "ALP custom messages FileNotFoundException for: "+ customsgsFile);
            } catch (Exception e) {
                logmsg(world, "ALP custom messages cannot be processed: "+ e.getClass().getSimpleName()+":"+e.getMessage());
            }
        }else{
            logmsg(world, "ALP custom messages not specified in World ATTRLIST...");
        }        
        
        logmsg(world, "ALP Custom post-init processing of "+ _objs.size()+ " objects commencing.");
               
        //for each item:...,preadj=the,adjs=nice sharp,noun=scissors,pronoun=it,plural=scissors
        for(int i=0;i<_objs.size();i++){
            
            obj = (AdvObject) _objs.elementAt(i);
            logmsg(world, "ALP processing "+ obj.getId());

            noun=obj.varGet(NOUN_LC);
            String nn = noun.strVal();
            if(nn.equals("")) {
                noun = obj.varGet(TYPE_LC);
                nn = noun.strVal();
            }
            if(! nn.equals("")) {
                nn = nn.toLowerCase();
                logmsg(world, "   noun:"+ noun.strVal());
                nouns.put(nn, nn);
            }
            boolean nnvalid = !nn.equals("");
            
            preadj=obj.varGet(PREADJ_LC);
            String pa = preadj.strVal();
            if(! pa.equals("")) {
                pa = pa.toLowerCase();
                logmsg(world, "   preadj:"+ pa);
                preAdjectives.put(pa,pa);
                if(nnvalid) preAdjNouns.put(nn, pa);
            }
            
            adjs=obj.varGet(ADJS_LC);
            String a = adjs.strVal();
            if(!a.equals("")) {
                a = a.toLowerCase();
                logmsg(world, "   adjs:"+ a);
                parseAdjs(a);
            }
            
            pronoun=obj.varGet(PRONOUN_LC);
            String pn = pronoun.strVal();                  
            if(!pn.equals("")) {
                pn = pn.toLowerCase();
                logmsg(world, "   pronoun:"+ pn);
                pronouns.put(pn, pn);
                if(nnvalid) pronounNouns.put(nn, pn);
            }
            
            plural=obj.varGet(PLURAL_LC);
            String pl = plural.strVal();
            if(!pl.equals("")){
                pl = pl.toLowerCase();
                logmsg(world, "   plural:"+ pl);
                plurals.put(pl, nn);
                if(nnvalid) {
                    pluralNouns.put(pl, nn);
                    nounPlurals.put(nn, pl);
                }
            }
        }
        
        // retrieve World Language SETS
        logmsg(world, "ALP processing Verbs...");        
        //what happens if no verbs have been defined? A NullPointerException!
        Dict vs = world.varGet("arrVerbs").dictVal();
        if( vs != null){
            for(int i = 0; i < vs.size(); i++){
                String v = ((Token)vs.elementAt(i)).strVal().toLowerCase();            
                logmsg(world, "   processing: "+ v);            
                verbs.put(v, v);      // the standard verb itself
                cmdAliases.put(v, v); // is also an alias for itself
                
                // cmdAliases ex: arrVerb_use        
                Dict alias = world.varGet("alias_"+v).dictVal();
                if( alias != null){
                    if(alias.size()> 0){
                        logmsg(world, "   processing '"+ v +"' aliases...");            
                        for (int ia = 0; ia < alias.size(); ia++){
                            String a = ((Token)alias.elementAt(ia)).strVal().toLowerCase();
                            if(!a.equalsIgnoreCase(v)){
                                logmsg(world, "               "+ a );            
                                verbs.put(a, a);
                                cmdAliases.put(a, v);
                            }
                        }
                    }
                }
            } 
        }
        logmsg(world, "ALP processing AltSay...");
        Dict alts = world.varGet("arrAltSay").dictVal();
        if( alts != null){
            for(int i = 0; i < alts.size(); i++){
                String w = ((Token)alts.elementAt(i)).strVal().toLowerCase();            
                logmsg(world, "   processing: "+ w);            
                say2.put(w, w);
            }
        }
        logmsg(world, "ALP processing AltUse...");
        Dict altu = world.varGet("arrAltUse").dictVal();
        if( altu != null){
            for(int i = 0; i < altu.size(); i++){
                String w = ((Token)altu.elementAt(i)).strVal().toLowerCase();            
                logmsg(world, "   processing: "+ w);            
                use2.put(w, w);
            }
        }
        logmsg(world, "ALP processing Withs...");
        Dict ws = world.varGet("arrWiths").dictVal();
        if( ws != null){
            for(int i = 0; i < ws.size(); i++){
                String w = ((Token)ws.elementAt(i)).strVal().toLowerCase();            
                logmsg(world, "   processing: "+ w);            
                withs.put(w, w);
            }
        }        
        logmsg(world, "ALP processing Ands...");
        Dict as = world.varGet("arrAnds").dictVal();
        if( as != null){
            for(int i = 0; i < as.size(); i++){
                String a = ((Token)as.elementAt(i)).strVal().toLowerCase();            
                logmsg(world, "   processing: "+ a);            
                ands.put(a, a);
            }
        }
                        
        logmsg(world, "Custom post-init processing complete.");
        return true;
    }    
    
    /** 
     * Part of world initialization.
     * store space-separated adjectives for an item into the adjectives dictionary.
     * @param adjs
     */
    void parseAdjs(String adjs){
        String[] _adjs = adjs.split(space);
        for(int i=0;i<_adjs.length;i++) adjectives.put(_adjs[i], _adjs[i]);                    
    }

    String listAliases(World world, String aliasSet){
        Dict alias = world.varGet(aliasSet).dictVal();
        String msg = ((Token)cmsgs.get("help_alias_prefix")).strVal()+ space;
        if( alias != null){
            if(alias.size()> 0){
                msg += HS_STRING;
                for (int ia = 0; ia < alias.size(); ia++){
                    if(ia > 0) msg +=", ";
                    msg += ((Token)alias.elementAt(ia)).strVal();
                }
                return msg += HE_STRING;
            }
        }
        return ((Token)cmsgs.get("help_alias_none")).strVal();
    }

    /**
     * Extract and hilite the message defined by msgkey
     * @param msgkey
     * @return text of message hilited as appropriate
     */
    protected String HL(String msgkey){
        try{
            String amsg = ((Token)cmsgs.get(msgkey)).strVal();
            amsg = amsg.replace(HS_MARK, HS_STRING);
            return amsg.replace(HE_MARK, HE_STRING);
        }catch(Exception x){
            return msgkey;
        }
    }
    
    protected Vector cmdHELP(multiplayer server, World world, Vector utils, Player player){
        logmsg(world,"ALP processing "+ ((Token)cmsgs.get("command_help")).strVal() +" command...");
        String arg1 = (String)utils.get(1);
        //cmd = "help";
        String sep =comma + space;
        String cr = "\n";
        String openopt = space+open_bracket;
        String closeopt = close_bracket;
        String lc = arg1.toLowerCase();
        String msg = "";
        if(lc.indexOf(space + ((Token)cmsgs.get("command_close")).strVal().toLowerCase()) > 0){
            msg = HL("help_close_insert");
            msg+= HL("help_close_example");
            msg+= listAliases(world, "alias_close");
        }
        else if(lc.indexOf(space + ((Token)cmsgs.get("command_drop")).strVal().toLowerCase()) > 0){
            msg = HL("help_drop_insert");
            msg+= HL("help_drop_example");
            msg+= listAliases(world, "alias_drop");
        }
        else if(lc.indexOf(space + ((Token)cmsgs.get("command_give")).strVal().toLowerCase()) > 0){
            msg = HL("help_give_insert");
            msg+= HL("help_give_example");
            msg+= listAliases(world, "alias_give");
        }
        else if(lc.indexOf(space + ((Token)cmsgs.get("command_go")).strVal().toLowerCase()) > 0){
            msg = HL("help_go_insert");
            msg+= HL("help_go_shortcuts");
            msg+= HL("help_go_example");
            msg+= listAliases(world, "alias_go");
        }
        else if(lc.indexOf(space + ((Token)cmsgs.get("command_look")).strVal().toLowerCase()) > 0){
            msg = HL("help_look_insert");
            msg+= HL("help_look_shortcuts");
            msg+= HL("help_look_example");
            msg+= listAliases(world, "alias_look");
        }
        else if(lc.indexOf(space + ((Token)cmsgs.get("command_turn")).strVal().toLowerCase()) > 0){
            msg = HL("help_turn_insert");
            msg+= HL("help_turn_shortcuts");
            msg+= HL("help_turn_example");
            msg+= listAliases(world, "alias_turn");
        }
        else if(lc.indexOf(space + ((Token)cmsgs.get("command_reverse")).strVal().toLowerCase()) > 0){
            msg = HL("help_reverse_insert");
            msg+= HL("help_reverse_example");
            msg+= listAliases(world, "alias_reverse");
        }
        else if(lc.indexOf(space + ((Token)cmsgs.get("command_open")).strVal().toLowerCase()) > 0){
            msg = HL("help_open_insert");
            msg+= HL("help_open_example");
            msg+= listAliases(world, "alias_open");
        }
        else if(lc.indexOf(space + ((Token)cmsgs.get("command_pick")).strVal().toLowerCase()) > 0){
            msg = HL("help_pick_insert");
            msg+= HL("help_pick_example");
            msg+= listAliases(world, "alias_pick");
        }
        else if(lc.indexOf(space + ((Token)cmsgs.get("command_search")).strVal().toLowerCase()) > 0){
            msg = HL("help_search_insert");
            msg+= HL("help_search_example");
            msg+= listAliases(world, "alias_search");
        }
        else if(lc.indexOf(space + ((Token)cmsgs.get("command_hide")).strVal().toLowerCase()) > 0){
            msg = HL("help_hide_insert");
            msg+= HL("help_hide_example");
            msg+= listAliases(world, "alias_hide");
        }
        else if(lc.indexOf(space + ((Token)cmsgs.get("command_put")).strVal().toLowerCase()) > 0){
            msg = HL("help_put_insert");
            msg+= HL("help_put_example");
            msg+= listAliases(world, "alias_put");
        }
        else if(lc.indexOf(space + ((Token)cmsgs.get("command_save")).strVal().toLowerCase()) > 0){
            if(lc.indexOf(space + ((Token)cmsgs.get("command_exit")).strVal().toLowerCase()) > 0){
                msg = HL("help_saveexit_insert");
                msg+= HL("help_saveexit_example");
            }else{
                msg = HL("help_save_insert");
                msg+= HL("help_save_example");
            }
        }
        else if(lc.indexOf(space + ((Token)cmsgs.get("command_enter")).strVal().toLowerCase()) > 0){
            msg = HL("help_enter_insert");
            msg+= HL("help_enter_example");
            msg+= listAliases(world, "alias_enter");
        }
        else if(lc.indexOf(space + ((Token)cmsgs.get("command_exit")).strVal().toLowerCase()) > 0){
            msg = HL("help_exit_insert");
            msg+= HL("help_exit_example");
            msg+= listAliases(world, "alias_exit");
        }
        else if(lc.indexOf(space + ((Token)cmsgs.get("command_use")).strVal().toLowerCase()) > 0){
            msg = HL("help_use_insert");
            msg+= HL("help_use_example");
            msg+= listAliases(world, "alias_use");
            msg+= HL("help_use_alt_syntax");
            for(int i=0;i<use2.size();i++){
                if(i>0&&i<use2.size()+1)
                    msg +=", ";
                msg += use2.elementAt(i);
            }
            msg+= HL("help_use_alt_example");            
        }
        else if(lc.indexOf(space + ((Token)cmsgs.get("command_say")).strVal().toLowerCase()) > 0){
            msg = HL("help_say_insert");
            msg+= HL("help_say_example");
            msg+= listAliases(world, "alias_say");
            msg+= HL("help_say_alt_syntax");
            for(int i=0;i<say2.size();i++){
                if(i>0&&i<say2.size()+1)
                    msg +=", ";
                msg += say2.elementAt(i);
            }
            msg+= HL("help_say_alt_example");            
        }
        else if(lc.indexOf(space + ((Token)cmsgs.get("command_logout")).strVal().toLowerCase()) > 0){
            msg = HL("help_logout_insert");
            msg+= HL("help_logout_example");
        }
        else{
            msg = HL("help_intro");
            msg+= HL("help_ctrl_insert");
            msg+= HL("help_ctrl_insert2");
            msg+= HL("help_table_intro");
            msg+= HL("help_table");
            msg+= HL("help_table_close");        
        }
        player.getClient().display(msg); 
        return clearCmdVector();
    }
}
