package org.dimx;

import cleoni.adv.*;
import java.util.Vector;

/**
 * Still in-development. Not ready for production game servers.  
 * API still subject to significant change.
 * @author Carl Nagle
 */
public class LanguageParser implements ISayProcessor{

    public boolean logDebug = true; // enables extensive debug logging

    public static String comma = ",";
    public static String semicolon = ";";
    public static String colon = ":";
    public static String period = ".";
    public static String exclamation = "!";
    public static String question = "?";
    public static String space = " ";
    public static String open_bracket = "[";
    public static String close_bracket = "]";

    public static String GO_LC = "go";
    public static String PICK_LC = "pick";
    public static String DROP_LC = "drop";
    public static String LOOK_LC = "look";
    public static String L_LC = "l";
    public static String SEARCH_LC = "search";
    public static String OPEN_LC = "open";
    public static String CLOSE_LC = "close";
    public static String USE_LC = "use";
    public static String HIDE_LC = "hide";
    public static String GIVE_LC = "give";
    public static String PUT_LC = "put";
    public static String ROTATE_LC = "rotate";
    public static String SAY_LC = "say";
    public static String SAVE_LC = "save";
    public static String SAVEEXIT_LC = "savexit";
    public static String ENTER_LC = "enter";
    public static String EXIT_LC = "exit";
    public static String IN_LC = "in";
    public static String ON_LC = "on";
    public static String WITH_LC = "with";
    public static String TO_LC = "to";
    public static String N_LC = "n";
    public static String S_LC = "s";
    public static String E_LC = "e";
    public static String W_LC = "w";
    public static String U_LC = "u";
    public static String D_LC = "d";
    public static String N_UC = "N";
    public static String S_UC = "S";
    public static String E_UC = "E";
    public static String W_UC = "W";
    public static String U_UC = "U";
    public static String D_UC = "D";
    public static String TL_LC = "tl";
    public static String TR_LC = "tr";
    public static String TURN_LC = "turn";
    public static String RIGHT_LC = "right";
    public static String LEFT_LC = "left";
    public static String UP_LC = "up";
    public static String DOWN_LC = "down";
    public static String NORTH_LC = "north";
    public static String SOUTH_LC = "south";
    public static String EAST_LC = "east";
    public static String WEST_LC = "west";
    public static String SLASH_HELP = null;// "/help" -- after init LocalizedData 
    public static String SLASH_SAVE = null;// "/save" -- after init LocalizedData 
    public static String SLASH_SAVEEXIT = null;// "/save exit" -- after init LocalizedData 
    public static String SLASH_DISCONNECT = null;// "/disconnect" -- after init LocalizedData 

    /** localized predicates in, on, with, to **/
    public static String[] PREDICATES_LC = null;
    
    /**
     * initialize the localized data like predicate array and SLASH_HELP
     * Only happens once--at the time of the first attempt to parse a command
     * @param world
     */
    void initLocalizedData(World world){
        if(PREDICATES_LC == null){
            PREDICATES_LC = new String[]{
                /* in */  world.msgs.msg[30].toLowerCase(), 
                /* on */  world.msgs.msg[31].toLowerCase(),
                /* with */world.msgs.msg[32].toLowerCase(), 
                /* to */  world.msgs.msg[33].toLowerCase()
            };
            SLASH_HELP = "/"+ world.msgs.cmd[26];
            SLASH_SAVE = "/"+ world.msgs.cmd[24];
            SLASH_SAVEEXIT = SLASH_SAVE +space+ world.msgs.cmd[23].toLowerCase();
            SLASH_DISCONNECT = "/"+ world.msgs.cmd[2];
        }
    }
    
    /**
     * only logs to world debug logger if localDebug is true
     * @param world
     * @param message
     */
    void logmsg(World world, String message){
        if (logDebug) world.logger.debug(message);        
    }
    
    /** 
     * return the index within arg immediately following a found PREDICATE.
     * this represents the start of the predicate object text.
     * the index will be -1 if no predicate was found or if no additional 
     * text exists after the actual predicate word (in on with to).
     **/
    int indexAfterPredicate(World world, String arg_lc, int startindex){
        if (startindex < 0) startindex = 0;
        if (arg_lc == null) return -1;
        if (startindex >= arg_lc.length()) return -1;
        int l = 0;
        int i = -1;
        String pred = null;
        for(;l < PREDICATES_LC.length; l++){
            pred = space + PREDICATES_LC[l]+ space;
            i = arg_lc.indexOf(pred);
            if (i == -1) continue;
            i += pred.length()-1;
            logmsg(world, "LP found PREDICATE '"+ pred +"'");
            return i < arg_lc.length() ? i : -1;
        }
        return -1;
    }
    
    /** 
     * return the index within arg immediately following a found PREDICATE.
     * this represents the start of the predicate object text.
     * the index will be -1 if no predicate was found or if no additional 
     * text exists after the actual predicate word (in on with to).
     **/
    int indexAfterPredicate(World world, String arg_lc){
        return indexAfterPredicate(world, arg_lc, 0);
    }
    
    /** 
     * return the index within arg of the *last* PREDICATE.
     * this represents the end of the text prior to the last predicate.
     * the index will be -1 if no predicate was found or if no additional 
     * text exists after the last predicate word (in on with to).
     **/
    int indexOfLastPredicate(World world, String arg_lc){
        int lastindex = -1;
        int tryindex = 0;
        int index = 0;        
        String pred = null;
        while(index != -1){
            for(int l = 0;l < PREDICATES_LC.length; l++){
                pred = space + PREDICATES_LC[l]+ space;
                index = arg_lc.indexOf(pred, tryindex);
                if (index == -1) continue;
                lastindex = index;
                tryindex = index+ pred.length()-1;
                break; //try again until index == -1
            }
        }
        return lastindex;
    }
    
    /**
     * Display ALL player SAY messages and chat in the Message panel.
     * 
     * If the player selects another player for the SAY then that player will 
     * receive the message but the selected player will NOT receiver player focus.
     * 
     * If the player sends a generic message with [ENTER], or selects a non-player 
     * object then normal SAY processing occurs in addition to displaying the text 
     * in the player's message panel.
     * 
     * The API is subject to change as the parent Interface becomes fully developed.
     * 
     * @param server
     * @param world
     * @param cmd
     * @param arg1
     * @param arg2
     * @param player
     * @return NULL to allow normal SAY processing to continue; 
     *         Vector{new cmd,arg1,arg2} to perform standard engine cmd.
     *         Vector{"","",""} to indicate all processing complete already.
     */
    public Vector processSay(multiplayer server, World world, String cmd, String arg1, String arg2, Player player) {
        try{ 
            // ALWAYS display my chat in message panel
            logmsg(world, "LP: processing '"+ arg1 +"'");
            return parseCommands(server, world, cmd, arg1, arg2, player);
        }
        catch(Exception x){
            logmsg(world, "LP: processSay Exception '"+ x.getClass().getSimpleName()+":"+x.getMessage() +"'");
        }
        return null;
    }
    
    /**
     * Enhanced Player Chat support in the message window.
     * This overrides normal game engine chat handling.
     * @param server
     * @param world
     * @param utils
     * @param player
     * @return NULL to allow normal SAY processing to continue; 
     *         Vector{"","",""} to bypass normal SAY processing.
     */
    public Vector cmdSAY(multiplayer server, World world, Vector utils, Player player){
        try{
            //cmd = "say";
            AdvObject target = world.getObject((String)utils.get(2));
            if(target != null){
                if(target.isaCharacter()){
                    player.getClient().receive(new Message(player, player.getId(), (String)utils.get(1)));
                    //talk to a player without changing focus
                    target.hear(player, (String)utils.get(1));
                    return clearCmdVector();                
                }
            }            
        }catch(Exception x){
            logmsg(world, "LP: cmdSAY Exception '"+ x.getClass().getSimpleName()+":"+x.getMessage() +"'");
        }
        return null;        
    }

    /** SAY something TO someone  processing*/
    protected Vector cmdSAYTO(multiplayer server, World world, Vector utils, Player player){
        logmsg(world,"LP processing SAY something TO someone");
        String arg1 = (String)utils.get(1);
        String arg_lc = arg1.toLowerCase();
        String verb = world.msgs.cmd[14].toLowerCase()+space;
        int sayindex = arg_lc.indexOf(verb); //"say "
        sayindex += verb.length(); //to start of actual say text
        int predindex = indexOfLastPredicate(world, arg_lc);
        int toindex = indexAfterPredicate(world, arg_lc, predindex);
        
        String something = arg1.substring(sayindex).trim();
        
        // if not valid display 'don't understand' message
        if (((predindex - sayindex) < 2 )||(toindex == -1)) {
            logmsg(world,"LP SAY '"+ something +"'input specifies no target. Reverting to normal say output.");
            utils.set(0, SAY_LC);
            utils.set(1, something);
            utils.set(2, "");
            return utils;
        }
        
        something = arg1.substring(sayindex, predindex).trim();
        String someone = arg1.substring(toindex).trim();
        logmsg(world,"LP SAY '"+ something +"' TO '"+ someone+"'");
        
        String[] ids = getMatchingItemIds(world, player, someone.toLowerCase());
        if(ids.length==0) {// not found
            logmsg(world,"LP SAY target '"+ someone +"' does not exist. Reverting to normal say output.");
            utils.set(0, SAY_LC);
            utils.set(1, something);
            utils.set(2, "");
            return utils;
        }
        utils.set(0, SAY_LC);
        utils.set(1, something);
        utils.set(2, ids[0]);
        // new CHAT version of Say
        return cmdSAY(server, world, utils, player);
    }    
    
    /**
     * Search command processing.
     * Attempts to match arg1 with local items and inventory.
     * Multiple matches will always default to the first one found.
     * If no match is found the normal engine tries to search the player's container.
     * @param server
     * @param world
     * @param utils
     * @param player
     * @return
     */
    protected Vector cmdSEARCH(multiplayer server, World world, Vector utils, Player player){
        logmsg(world,"LP processing SEARCH...");
        String arg1 = (String)utils.get(1);
        String[] ids = getMatchingItemIds(world, player, arg1.toLowerCase());
        utils.set(0, SEARCH_LC);
        if(ids.length==0) // not found
            utils.set(1, "");
        else
            utils.set(1, ids[0]);
        return utils;
    }    

    /**
     * Currently just displays localized "I could not do that"
     * @param world
     * @param player
     * @param arg1
     * @return clearCmdVector()
     */
    protected Vector badOPEN(World world, Player player, String arg1){
        logmsg(world, "LP processing bad OPEN...");
        player.getClient().display(world.msgs.msg[141]); // I could not do that
        return clearCmdVector();
    }
    
    /**
     * Open command processing.
     * Attempts to match arg1 with local items and inventory.
     * Multiple matches will always default to the first one found.
     * @param server
     * @param world
     * @param utils
     * @param player
     * @return
     */
    protected Vector cmdOPEN(multiplayer server, World world, Vector utils, Player player){
        logmsg(world,"LP processing OPEN...");
        String arg1 = (String)utils.get(1);
        String[] ids = getMatchingItemIds(world, player, arg1.toLowerCase());
        utils.set(0, OPEN_LC);
        if(ids.length==0) {// not found
            return badOPEN(world, player, arg1);
        }
        else{
            utils.set(1, ids[0]);
        }
        return utils;
    }    

    /**
     * Currently just displays localized "I could not do that"
     * @param world
     * @param player
     * @param arg1
     * @return clearCmdVector()
     */
    protected Vector badCLOSE(World world, Player player, String arg1){
        logmsg(world,"LP processing bad CLOSE...");
        player.getClient().display(world.msgs.msg[141]); // I could not do that
        return clearCmdVector();
    }
    
    /**
     * Close command processing.
     * Attempts to match arg1 with local items and inventory.
     * Multiple matches will always default to the first one found.
     * @param server
     * @param world
     * @param utils
     * @param player
     * @return
     */
    protected Vector cmdCLOSE(multiplayer server, World world, Vector utils, Player player){
        logmsg(world,"LP processing CLOSE...");
        String arg1 = (String)utils.get(1);
        String[] ids = getMatchingItemIds(world, player, arg1.toLowerCase());
        boolean result = false;
        utils.set(0,CLOSE_LC);
        if(ids.length==0) {// not found
            return badCLOSE(world, player, arg1);
        }
        else{
            utils.set(1, ids[0]);
        }
        return utils;
    }    

    /**
     * Currently just displays localized "I could not do that"
     * @param world
     * @param player
     * @param arg1
     * @return clearCmdVector()
     */
    protected Vector badHIDE(World world, Player player, String arg1){
        logmsg(world,"LP processing bad HIDE...");
        player.getClient().display(world.msgs.msg[141]); // I could not do that
        return clearCmdVector();
    }

    /**
     * Hide command processing.
     * Attempts to match arg1 with local items and inventory.
     * Multiple matches will always default to the first one found.
     * @param server
     * @param world
     * @param utils
     * @param player
     * @return
     */
    protected Vector cmdHIDE(multiplayer server, World world, Vector utils, Player player){
        logmsg(world,"LP processing HIDE...");
        String arg1 = (String)utils.get(1);
        String[] ids = getMatchingItemIds(world, player, arg1.toLowerCase());
        utils.set(0,HIDE_LC);
        if(ids.length==0) {// not found
            return badHIDE(world, player, arg1);
        }
        else{
            utils.set(1,ids[0]);
        }
        return utils;
    }

    /**
     * Currently just displays localized "I could not do that"
     * @param world
     * @param player
     * @param arg1
     * @return clearCmdVector()
     */
/*    protected Vector badLOOK(World world, Player player, String arg1){
        logmsg(world,"LP processing bad LOOK...");
        player.getClient().display(world.msgs.msg[141]); // I could not do that
        return clearCmdVector();
    }
*/
    /**
     * Look command processing.
     * Attempts to match arg1 with local items and inventory.
     * Multiple matches will always default to the first one found.
     * If no matches are found the normal engine tries to look the current container.
     * @param server
     * @param world
     * @param utils 
     * @param player
     * @return
     */
    protected Vector cmdLOOK(multiplayer server, World world, Vector utils, Player player){
        logmsg(world,"LP processing LOOK...");
        String arg1 = (String)utils.get(1);
        String[] ids = getMatchingItemIds(world, player, arg1.toLowerCase());
        utils.set(0,LOOK_LC);
        if(ids.length > 0) {
            utils.set(1,ids[0]);
        }else{
            utils.set(1,"");
        }
        return utils;
    }
    
    /**
     * Reverse command processing.
     * @param server
     * @param world
     * @param utils 
     * @param player
     * @return
     */
    protected Vector cmdREVERSE(multiplayer server, World world, Vector utils, Player player){
        logmsg(world,"LP processing REVERSE...");
        utils.set(0,ROTATE_LC);
        utils.set(1,"b");
        return utils;
    }
    
    /**
     * Currently just displays localized "Which way?"
     * @param server
     * @param world
     * @param player
     * @param direction
     * @return clearCmdVector()
     */
    protected Vector badGO(multiplayer server, World world, Player player, String direction){
        logmsg(world,"LP processing bad GO input...");
        player.getClient().display(world.msgs.msg[178]);
        return clearCmdVector(); //no further cmd to execute
    }
    
    /**
     * Go command processing.
     * @param server
     * @param world
     * @param utils -- arg1 -- Must be a valid WAY MARKER as used in Room Links:  
     * Valid values are "N", "S", "E", "W", "U", "D"
     * @param player
     * @return
     */
    public Vector cmdGO(multiplayer server, World world, Vector utils, Player player) throws DimxException{
        //logger.debug("Prima sono in: " + thisPlayer.getRoom());
        logmsg(world,"LP processing GO...");
        String arg1 = (String)utils.get(1);
        utils.set(0, GO_LC);
        utils.set(1, ""); //default to unknown direction
        Room room = player.getRoom();
        if ((room != null) && (!room.equals(""))) {
            Dict links = room.getLinks();
            if(links == null) {
                return badGO(server, world, player, arg1);
            }else{
                java.util.Enumeration list = links.elements();
                Link link = null;
                String dir = null;
                boolean found = false;
                while(list.hasMoreElements()&& !found){                                
                    link = (Link) list.nextElement();
                    dir = link.getDirection(room.getId());
                    found = dir.equalsIgnoreCase(arg1);
                }
                if(found){
                    utils.set(1,link.getId());
                }else{
                    return badGO(server, world, player, arg1);
                }
            }
        }
        //return badGO(server, world, player, arg1);
        return utils;
    }

    /**
     * Turn Right, Turn Left command processing.
     * @param server
     * @param world
     * @param utils 
     * @param player
     * @return
     */
    public Vector cmdTURN(multiplayer server, World world, Vector utils, Player player){
        logmsg(world,"LP processing TURN LEFT/RIGHT...");
        String left = "l";
        String arg1 = (String)utils.get(1);
        utils.set(0,"rotate");
        utils.set(1,"r");
        boolean result = false;
        if(arg1.equals(TR_LC)){
            ; // already set "r"
        }else if(arg1.equals(TL_LC)){
            utils.set(1,left);
        }else {
            String turnleft = world.msgs.msg[173].toLowerCase();
            String arglc = arg1.toLowerCase();
            String[] leftwords = turnleft.split(space);
            int index = 0;
            int last = -1;
            int i = 0;
            if(leftwords.length > 1){
                for(; i <leftwords.length;i++){
                    try{
                        index = arglc.indexOf(leftwords[i], last+1);                    
                        if(index > last) {
                            last = index;
                            continue;
                        }
                    }catch(Exception e){}
                    break;
                }
                if((index > last)&&(i == leftwords.length)) utils.set(1,left);
            }            
        }
        return utils;
    }

    /**
     * Currently just displays localized "I could not do that"
     * @param world
     * @param player
     * @param arg1
     * @return clearCmdVector()
     */
/*    protected Vector badPICK(World world, Player player, String arg1){
        logmsg(world,"LP processing bad PICK...");
        player.getClient().display(world.msgs.msg[141]); // I could not do that
        return clearCmdVector();
    }
*/
    /**
     * Pick command processing.
     * Attempts to match arg1 with local items and inventory.
     * Multiple matches will always default to the first one found.
     * If no matches are found the normal engine tries to pick up ALL nearby items it can.
     * @param server
     * @param world
     * @param utils
     * @param player
     * @return
     */
    public Vector cmdPICK(multiplayer server, World world, Vector utils, Player player){
        logmsg(world,"LP processing PICK...");
        String arg1 = (String)utils.get(1);
        String[] ids = getMatchingContainerIds(world, player.container, arg1.toLowerCase());
        utils.set(0,PICK_LC);
        logmsg(world,"LP PICK matches "+ ids.length +" items...defaulting to first item.");            
        if(ids.length==0) // not found
            // if we send no object the game actually tries to PICK *everything* in the container!
            //return badPICK(world, player, aid, arg1);        
            utils.set(1,"");        
        else
            utils.set(1,ids[0]);        
        return utils;
    }

    /**
     * Currently just displays localized "I could not do that"
     * @param world
     * @param player
     * @param arg1
     * @return clearCmdVector()
     */
    protected Vector badDROP(World world, Player player, String arg1){
        logmsg(world,"LP processing bad DROP...");
        player.getClient().display(world.msgs.msg[141]); // I could not do that
        return clearCmdVector();
    }

    /**
     * Drop command processing.
     * Attempts to match arg1 with local items and inventory.
     * Multiple matches will always default to the first one found.
     * @param server
     * @param world
     * @param utils
     * @param player
     * @return
     */
    public Vector cmdDROP(multiplayer server, World world, Vector utils, Player player){
        logmsg(world,"LP processing DROP...");
        String arg1 = (String)utils.get(1);
        utils.set(0,DROP_LC);
        String[] ids = getMatchingInventoryIds(world, player, arg1.toLowerCase());
        if(ids.length==0) {// not found
            return badDROP(world, player, arg1);        
        }else{
            utils.set(1, ids[0]);
            if(ids.length > 1)
                logmsg(world,"LP DROP matches multiple items...defaulting to first item.");            
        }
        return utils;
    }

    /**
     * Currently just displays localized "I could not do that"
     * @param world
     * @param player
     * @param arg1
     * @return clearCmdVector()
     */
    protected Vector badGIVE(World world, Player player, String arg1){
        logmsg(world,"LP processing bad GIVE...");
        player.getClient().display(world.msgs.msg[141]); // I could not do that
        return clearCmdVector();
    }

    /**
     * Give command processing.
     * Attempts to match objtext with local items and inventory.
     * Attempts to match predtext with local items and inventory.
     * Multiple matches on either will always default to the first one found.
     * @param server
     * @param world
     * @param utils
     * @param player
     * @param objtext -- text left of any command predicate (in, on, with, to)
     * @param predtext -- text right of any command predicate
     * @return
     */
    protected Vector cmdGIVE(multiplayer server, World world, Vector utils, Player player, 
                             String objtext, String predtext){
        logmsg(world,"LP processing GIVE...");
        String arg1 = (String)utils.get(1);
        utils.set(0,GIVE_LC);
        String[] ids = getMatchingItemIds(world, player, objtext.toLowerCase());
        if(ids.length==0) {// not found
            return badGIVE(world, player, arg1);
        }else{
            utils.set(1,ids[0]);
            ids = getMatchingItemIds(world, player, predtext.toLowerCase());
            if(ids.length == 0){
                return badGIVE(world, player, arg1);
            }else{
                utils.set(2,ids[0]);
            }
        }
        return utils;
    }

    /**
     * Currently just displays localized "I could not do that"
     * @param world
     * @param player
     * @param arg1
     * @return clearCmdVector()
     */
    protected Vector badPUT(World world, Player player, String arg1){
        logmsg(world,"LP processing bad PUT...");
        player.getClient().display(world.msgs.msg[141]); // I could not do that
        return clearCmdVector();
    }
    
    /**
     * Put command processing.
     * Attempts to match objtext with local items and inventory.
     * Attempts to match predtext with local items and inventory.
     * Multiple matches on either will always default to the first one found.
     * @param server
     * @param world
     * @param utils
     * @param player
     * @param objtext -- text left of any command predicate (in, on, with, to)
     * @param predtext -- text right of any command predicate
     * @return
     */
    protected Vector cmdPUT(multiplayer server, World world, Vector utils, 
                             Player player, String objtext, String predtext){
        logmsg(world,"LP processing PUT...");
        utils.set(0,PUT_LC);
        String arg1 = (String)utils.get(1);
        String[] ids = getMatchingItemIds(world, player, objtext.toLowerCase());
        if(ids.length==0) {// not found
            return badPUT(world, player, arg1);
        }else{
            utils.set(1,ids[0]);
            ids = getMatchingItemIds(world, player, predtext.toLowerCase());
            if(ids.length == 0){
                return badPUT(world, player, arg1);
            }else{
                utils.set(2,ids[0]);
            }
        }
        return utils;
    }

    /**
     * Currently just displays localized "I could not do that"
     * @param world
     * @param player
     * @param arg1
     * @param objtext
     * @param predtext
     * @return clearCmdVector()
     */
    protected Vector badUSE(World world, Player player, String arg1, String objtext, String predtext){
        logmsg(world,"LP processing bad USE OBJ: '"+ objtext +"', PRED: '"+ predtext +"'");
        player.getClient().display(world.msgs.msg[141]); // I could not do that
        return clearCmdVector();
    }

    /**
     * Use and Use With command processing.
     * Attempts to match objtext with local items and inventory.
     * Attempts to match predtext with local items, inventory, and room links.
     * Multiple matches on either will always default to the first one found.
     * @param server
     * @param world
     * @param utils
     * @param player
     * @param objtext -- text left of any command predicate (in, on, with, to)
     * @param predtext -- text right of any command predicate
     * @return
     */
    protected Vector cmdUSE(multiplayer server, World world, Vector utils, Player player, 
                             String objtext, String predtext){
        logmsg(world,"LP processing USE...");
        String arg1 = (String)utils.get(1);
        utils.set(0,USE_LC);
        String[] ids = getMatchingItemIds(world, player, objtext.toLowerCase());
        if(ids.length==0){ // not found
            return badUSE(world, player, arg1, objtext, predtext);
        }else{
            utils.set(1,ids[0]);
            if(predtext.length() > 0){
                ids = getMatchingItemIds(world, player, predtext.toLowerCase());
                if(ids.length == 0){
                    ids = getMatchingLinks(world, player, predtext.toLowerCase());
                    if(ids.length > 1)
                        logmsg(world, "LP USE on Link multiple matches...using first");
                }
                if(ids.length == 0)                    
                    utils.set(2,"");// we have an unmatched predicate
                else{
                    utils.set(0,"use2");
                    utils.set(2,ids[0]);            
                }
            }
        }
        return utils;
    }

    /**
     * Currently just displays localized "I could not do that"
     * @param world
     * @param player
     * @param arg1
     * @return clearCmdVector()
     */
/*    protected Vector badEXIT(World world, Player player, String arg1){
        logmsg(world,"LP processing bad EXIT...");
        player.getClient().display(world.msgs.msg[141]); // I could not do that
        return clearCmdVector();
    }
*/
    /**
     * Exit command processing.
     * @param server
     * @param world
     * @param utils
     * @param player
     * @return
     */
    protected Vector cmdEXIT(multiplayer server, World world, Vector utils, Player player){
        logmsg(world,"LP processing EXIT...");
        String arg1 = (String)utils.get(1);
        utils.set(0,EXIT_LC);
        utils.set(1,"");
        return utils;
    }
    
    /**
     * Currently just displays localized "I could not do that"
     * @param world
     * @param player
     * @param arg1
     * @return clearCmdVector()
     */
/*    protected Vector badDISCONNECT(World world, Player player, String arg1){
        logmsg(world,"LP processing bad DISCONNECT...");
        player.getClient().display(world.msgs.msg[141]); // I could not do that
        return clearCmdVector();
    }
*/
    /**
     * Disconnect command processing.
     * @param server
     * @param world
     * @param utils
     * @param player
     * @return
     */
    protected Vector cmdDISCONNECT(multiplayer server, World world, Vector utils, Player player){
        logmsg(world,"LP processing DISCONNECT...");
        utils.set(0,"logout");
        utils.set(1,"");
        return utils;
    }
    
    /**
     * Currently just displays localized "I could not do that"
     * @param world
     * @param player
     * @param arg1
     * @return clearCmdVector()
     */
/*    protected Vector badSAVE(World world, Player player, String arg1){
        logmsg(world,"LP processing bad SAVE...");
        player.getClient().display(world.msgs.msg[141]); // I could not do that
        return clearCmdVector();
    }
*/
    /**
     * SAVE and SAVEEXIT command processing.
     * @param server
     * @param world
     * @param utils
     * @param player
     * @return
     */
    protected Vector cmdSAVE(multiplayer server, World world, Vector utils, Player player){
        logmsg(world,"LP processing SAVE...");
        String arg1_lc = ((String)utils.get(1)).toLowerCase();
        utils.set(0,SAVE_LC);
        utils.set(1,"");
        String sexit = world.msgs.cmd[23].toLowerCase(); //exit
        if(arg1_lc.indexOf(sexit)> 0){
            utils.set(0,SAVEEXIT_LC);
        } 
        return utils;
    }
    
    /**
     * Currently just displays localized "I could not do that"
     * @param world
     * @param player
     * @param arg1
     * @return clearCmdVector()
     */
    protected Vector badENTER(World world, Player player, String arg1){
        logmsg(world,"LP processing bad ENTER...");
        player.getClient().display(world.msgs.msg[141]); // I could not do that
        return clearCmdVector();
    }

    /**
     * ENTER command processing.
     * Attempts to match arg1 with local items.
     * Multiple matches will always default to the first one found.
     * @param server
     * @param world
     * @param utils
     * @param player
     * @return
     */
    public Vector cmdENTER(multiplayer server, World world, Vector utils, Player player){
        logmsg(world,"LP processing ENTER...");
        String arg1 = (String)utils.get(1);
        utils.set(0,ENTER_LC);
        String[] ids = getMatchingContainerIds(world, player.container, arg1.toLowerCase());
        if(ids.length==0) {// not found
            //utils.set(1,"");
            return badENTER(world, player, arg1);        
        }else{
            if(ids.length > 1)
                logmsg(world,"LP ENTER matches multiple items...defaulting to first item.");            
            utils.set(1,ids[0]);
        }
        return utils;
    }

    /**
     * Produce an initialized Vector with all args (0,1,2) empty ("").
     * @return initialized Vector
     */
    protected Vector clearCmdVector(){
    	Vector v = new Vector(3);
    	v.add("");//cmd
    	v.add("");//arg1
    	v.add("");//arg2
    	return v;
    }
    
    /**
     * The workhorse routine to parse user input to execute game commands
     * @param server
     * @param world
     * @param cmd
     * @param arg1
     * @param arg2
     * @param player
     * @return NULL to allow normal SAY processing to continue; 
     *         Vector{new cmd,arg1,arg2} to bypass normal SAY processing.
     */
    Vector parseCommands (multiplayer server, World world, String cmd, String arg1, String arg2, Player player) throws DimxException{
        
        logmsg(world,"LP: processing input: "+ arg1);
        initLocalizedData(world);//will only do so once per world
        String arg1_lc = arg1.toLowerCase();
        Vector utils = clearCmdVector();
        utils.set(1, arg1); // the say input
        String [] words = arg1.split(space); //TODO: watch for punctuation (,.!?, etc.)
        int i = 0;
        String verb = null;
        player.getClient().cmdNewmsg(); // clear the txtBox on next refresh
        while(i < words.length){            
            verb = words[i++].toLowerCase();
            world.debug("LP: word="+ verb);
            if(verb.equalsIgnoreCase(world.msgs.cmd[20])) continue;// ignore go
            if(verb.equals(N_LC)||verb.equalsIgnoreCase(world.msgs.msg[167])){
                utils.set(1, N_UC);
                return cmdGO(server, world, utils, player);
            }
            if(verb.equals(S_LC)||verb.equalsIgnoreCase(world.msgs.msg[168])) {
                utils.set(1, S_UC);
                return cmdGO(server, world, utils, player);
            }
            if(verb.equals(E_LC)||verb.equalsIgnoreCase(world.msgs.msg[169])){
                utils.set(1, E_UC);
                return cmdGO(server, world, utils, player);
            }
            if(verb.equals(W_LC)||verb.equalsIgnoreCase(world.msgs.msg[170])){
                utils.set(1, W_UC);
                return cmdGO(server, world, utils, player);
            }
            if(verb.equals(U_LC)||verb.equalsIgnoreCase(world.msgs.msg[171])){
                utils.set(1, U_UC);
                return cmdGO(server, world, utils, player);
            }
            if(verb.equals(D_LC)||verb.equalsIgnoreCase(world.msgs.msg[172])){
                utils.set(1, D_UC);
                return cmdGO(server, world, utils, player);
            }
            if(verb.equals(TL_LC)||verb.equals(LEFT_LC)){
                utils.set(1, TL_LC);
                return cmdTURN(server, world, utils, player);
            }
            if(verb.equals(TR_LC)||verb.equals(RIGHT_LC)){
                utils.set(1, TR_LC);
                return cmdTURN(server, world, utils, player);
            }
            if(verb.equals(TURN_LC)||world.msgs.msg[173].toLowerCase().startsWith(verb)) /* turn */
                return cmdTURN(server, world, utils, player);            
            if(verb.equalsIgnoreCase(world.msgs.cmd[8])) /* Pick */
                return cmdPICK(server, world, utils, player);
            if(verb.equalsIgnoreCase(world.msgs.cmd[9])) /* Drop */
                return cmdDROP(server, world, utils, player);
            if(verb.equals(L_LC)||verb.equalsIgnoreCase(world.msgs.cmd[5])) /* l || Look */
                return cmdLOOK(server, world, utils, player);
            if(verb.equalsIgnoreCase(world.msgs.cmd[10])) /* Search */
                return cmdSEARCH(server, world, utils, player);
            if(verb.equalsIgnoreCase(world.msgs.cmd[6])) /* Open */
                return cmdOPEN(server, world, utils, player);
            if(verb.equalsIgnoreCase(world.msgs.cmd[7])) /* Close */
                return cmdCLOSE(server, world, utils, player);
            if(verb.equalsIgnoreCase(world.msgs.cmd[11])) /* Hide */
                return cmdHIDE(server, world, utils, player);
            if(verb.equalsIgnoreCase(world.msgs.cmd[22])) /* Enter */
                return cmdENTER(server, world, utils, player);
            if(verb.equalsIgnoreCase(world.msgs.cmd[23])) /* Exit */
                return cmdEXIT(server, world, utils, player);
            if(verb.equalsIgnoreCase(world.msgs.cmd[25])) /* Reverse */
                return cmdREVERSE(server, world, utils, player);
            if(verb.equalsIgnoreCase(world.msgs.cmd[14])) /* Say */
                return cmdSAYTO(server, world, utils, player);
            
            //engine commands /HELP, /SAVE, /SAVE EXIT, /DISCONNECT
            if(verb.equalsIgnoreCase(SLASH_HELP)) /* /Help */
                return cmdHELP(server, world, utils, player);
            if(verb.equalsIgnoreCase(SLASH_SAVE)) /* /Save and /Save exit */
                return cmdSAVE(server, world, utils, player);

            if(verb.equalsIgnoreCase(SLASH_DISCONNECT)) /* /Disconnect */
                return cmdDISCONNECT(server, world, utils, player);
            
            // handle 2-object commands
            int predStart = indexAfterPredicate(world, arg1_lc);
            String objtext = (predStart == -1) ? arg1 : arg1.substring(0,predStart);
            String predtext = (predStart == -1) ? "" : arg1.substring(predStart);

            if(verb.equalsIgnoreCase(world.msgs.cmd[12])) {  /* Put */
                if(predStart==-1) return badPUT(world, player, arg1);
                return cmdPUT(server, world, utils, player, objtext, predtext);
            }
            if(verb.equalsIgnoreCase(world.msgs.cmd[13])) {  /* Give */
                if(predStart==-1) return badGIVE(world, player, arg1);
                return cmdGIVE(server, world, utils, player, objtext, predtext);
            }
            if(verb.equalsIgnoreCase(world.msgs.cmd[3])) {  /* Use */
                return cmdUSE(server, world, utils, player, objtext, predtext);
            }
        }
        return null;    
    }

    protected Vector cmdHELP(multiplayer server, World world, Vector utils, Player player){
        logmsg(world,"LP processing /HELP command...");
        String arg1 = (String)utils.get(1);
        //cmd = "help";
        String sep =comma + space;
        String cr = "\n";
        String openopt = space+open_bracket;
        String closeopt = close_bracket;
        String lc = arg1.toLowerCase();
        String msg = SLASH_HELP +" "+ world.msgs.msg[17]+"\n"+ /* /help <command> */
            /* /save, /save exit, /disconnect */ SLASH_SAVE +sep+ SLASH_SAVEEXIT +sep+ SLASH_DISCONNECT +cr+cr+
            /*Supported Commands:*/world.msgs.msg[16] +cr+
            /*close*/world.msgs.cmd[7] +sep+ /*drop*/world.msgs.cmd[9] +sep+
            /*enter*/world.msgs.cmd[22] +sep+ /*exit*/world.msgs.cmd[23] +sep+
            /*give*/world.msgs.cmd[13] +sep+ /*go*/world.msgs.cmd[20] +sep+
            /*hide*/world.msgs.cmd[11] +sep+ /*look*/world.msgs.cmd[5] +sep+
            /*open*/world.msgs.cmd[6] +sep+ /*pick*/world.msgs.cmd[8] +sep+
            /*put*/world.msgs.cmd[12] +sep+ /*reverse*/world.msgs.cmd[25] + sep +
            /*say*/world.msgs.cmd[14] + sep+ /*search*/world.msgs.cmd[10] +sep+
            /*use*/world.msgs.cmd[3] +cr;
        
        if(lc.indexOf(space + world.msgs.cmd[7].toLowerCase()) > 0)
            /*close*/msg = arg1+cr+cr+ world.msgs.cmd[7]+ space +world.msgs.msg[34] +cr;
        else if(lc.indexOf(space + world.msgs.cmd[9].toLowerCase()) > 0)
            /*drop*/ msg = arg1+cr+cr+ world.msgs.cmd[9]+ space +world.msgs.msg[34] +cr;
        else if(lc.indexOf(space + world.msgs.cmd[13].toLowerCase()) > 0)
         /*give to*/ msg = arg1+cr+cr+ world.msgs.cmd[13]+ space +world.msgs.msg[34] + space +
                           world.msgs.msg[33]+ space + world.msgs.msg[34] +cr;
        else if(lc.indexOf(space + world.msgs.cmd[20].toLowerCase()) > 0)
            /*go*/ msg = arg1+cr+cr+ openopt+world.msgs.cmd[20]+ closeopt+ space + 
                         world.msgs.msg[167]+space+world.msgs.msg[168]+space+world.msgs.msg[169]+space+
                         world.msgs.msg[170]+space+world.msgs.msg[171]+space+world.msgs.msg[172]+ cr+
                         openopt+world.msgs.cmd[20]+ closeopt+ space +"n s e w u d\n";
        else if(lc.indexOf(space + world.msgs.cmd[11].toLowerCase()) > 0)
            /*hide*/ msg = arg1+cr+cr+ world.msgs.cmd[11]+ space +world.msgs.msg[34] +cr;
        else if(lc.indexOf(space + world.msgs.cmd[5].toLowerCase()) > 0)
            /*look*/ msg = arg1+cr+cr+ world.msgs.cmd[5]+ openopt +world.msgs.msg[34]+closeopt+cr +
                    L_LC + openopt +world.msgs.msg[34]+closeopt+cr;
        else if(lc.indexOf(space + world.msgs.cmd[6].toLowerCase()) > 0)
            /*open*/ msg = arg1+cr+cr+ world.msgs.cmd[6]+ space +world.msgs.msg[34] +cr;
        else if(lc.indexOf(space + world.msgs.cmd[8].toLowerCase()) > 0)
            /*pick*/ msg = arg1+cr+cr+ world.msgs.cmd[8]+ openopt +world.msgs.msg[34]+closeopt+cr;
        else if(lc.indexOf(space + world.msgs.cmd[10].toLowerCase()) > 0)
            /*search*/ msg = arg1+cr+cr+ world.msgs.cmd[10]+ openopt +world.msgs.msg[34]+closeopt+cr;
        else if(lc.indexOf(space + world.msgs.cmd[12].toLowerCase()) > 0)
         /*put in*/ msg = arg1+cr+cr+ world.msgs.cmd[12]+ space +world.msgs.msg[34] + space +
                           world.msgs.msg[30]+ space + world.msgs.msg[34] +cr;
        else if(lc.indexOf(space + world.msgs.cmd[22].toLowerCase()) > 0)
            /*enter*/ msg = arg1+cr+cr+ world.msgs.cmd[22]+ space +world.msgs.msg[34]+cr;
        else if(lc.indexOf(space + world.msgs.cmd[23].toLowerCase()) > 0)
            /*exit*/ msg = arg1+cr+cr+ world.msgs.cmd[23]+ cr;
        else if(lc.indexOf(space + world.msgs.cmd[25].toLowerCase()) > 0)
            /*reverse*/ msg = arg1+cr+cr+ world.msgs.cmd[25]+ sep +
            /*turn left,right*/world.msgs.msg[173]+ sep + world.msgs.msg[174]+ cr +
            /*left,right*/"Left, Right, tl, tr\n";
        else if(lc.indexOf(space + world.msgs.cmd[3].toLowerCase()) > 0)
         /*use*/ msg = arg1+cr+cr+ world.msgs.cmd[3]+ space +world.msgs.msg[34] + openopt +
                           world.msgs.msg[32]+ space + world.msgs.msg[34]+closeopt+cr;        
        else if(lc.indexOf(space + world.msgs.cmd[14].toLowerCase()) > 0)
         /*say to */ msg = arg1+cr+cr+ world.msgs.cmd[14] + " text " +
                           world.msgs.msg[33].toUpperCase()+ " character\n";        
        player.getClient().display(msg); 
        return clearCmdVector();
    }
    
    /**
     * Attempt to match the input_lc with the Name of an AdvObject in the items Dict.
     * Will attempt to id items in open container items, too.
     * @param world
     * @param input_lc
     * @param items
     * @return 0 or more item Ids who's full NAME can be found in text_lc
     */
    protected String[] getMatchingIds(World world, String input_lc, Dict items){
        Vector ids = new Vector();
        if(items.size()==0) return new String[0];        
        String id = null;
        AdvObject obj = null;
        String oname = null;
        Dict contents = null;
        for(int i = 0;i < items.size(); i++){
            obj = (AdvObject)items.elementAt(i);
            id = obj.getId();
            oname = obj.name.strVal().toLowerCase();
            //currently we only match on the item's name
            if(input_lc.indexOf(oname)>= 0)
                ids.add(id);
            // check for contents of open containers
            contents = obj.getContents(true);
            if(contents != null){
                String[] temp = getMatchingIds(world, input_lc, contents);
                for(int t = 0; t < temp.length;t++) ids.add(temp[t]);
            }
        }
        return (String[])ids.toArray(new String[ids.size()]);
    }
    
    /**
     * Attempt to match the input_lc with the Name of an item in the Player's inventory.
     * Internally, inventory contents are sought with player.getContents(false)
     * @param world
     * @param player
     * @param input_lc
     * @return 0 or more item Ids who's full NAME can be found in text_lc
     */
    protected String[] getMatchingInventoryIds(World world, Player player, String input_lc) {
        return getMatchingIds(world, input_lc, player.getContents(false));
    }

    /**
     * Attempt to match the input_lc with the Name of an item in the provided container.
     * Internally, container contents are sought with container.getContents(false)
     * @param world
     * @param container
     * @param input_lc
     * @return 0 or more item Ids who's full NAME can be found in input_lc
     */
    protected String[] getMatchingContainerIds(World world, AdvObject container, String input_lc) {
        return getMatchingIds(world, input_lc, container.getContents(false));
    }

    /**
     * Attempt to match the text_lc with the Name of a Link in the Player.Room.
     * @param world
     * @param player
     * @param text_lc
     * @return 0 or more Link Ids whose full NAME can be found in text_lc
     */
    String[] getMatchingLinks(World world, Player player, String text_lc){
        String[] ids = new String[0];
        Room room = player.getRoom();
        if ((room != null) && (!room.equals(""))) {
            Dict links = room.getLinks(); // can be null;
            if(links != null) ids = getMatchingIds(world, text_lc, links);
        }
        return ids;
    }
    
    /**
     * Attempt to match the text_lc with the Name of a World Item.
     * We try getMatchingContainerIds first then getMatchingInventoryIds if no 
     * container items were found.
     * @param world
     * @param player
     * @param text_lc
     * @return 0 or more item Ids whose full NAME can be found in text_lc
     */
    String[] getMatchingItemIds(World world, Player player, String text_lc){
        String[] ids = getMatchingContainerIds(world, player.container, text_lc);
        if(ids.length==0) // not found
            ids = getMatchingInventoryIds(world, player, text_lc);        
        if(ids.length > 1)
            logmsg(world,"LP getMatchingItemIds matches multiple items...defaulting to first item.");            
        return ids;
    }        
}
